import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.stream.Collectors;

public class DerivationTree implements Comparable<DerivationTree> {

	/**
	 * The grammar associated to this derivation tree.
	 */
	public final ContextFreeGrammar grammar;
	
	/**
	 * The symbol at the root of the tree. If this is a terminal symbol, the tree must not have children.
	 * Otherwise, there has to be a corresponding production in the grammar.
	 */
	public final Symbol symbol;
	
	/**
	 * The list of child trees.
	 */
	public final List<DerivationTree> children;

	public void check() {
		if (!grammar.alphabet.contains(symbol) && !grammar.nonTerminals.contains(symbol))
			throw new IllegalArgumentException("Not a valid symbol: " + symbol);
		if (children.isEmpty()) return;
		if (!(symbol instanceof NonTerminal))
			throw new IllegalArgumentException("Cannot make derivation of terminal: " + symbol);
		List<Symbol> rhs = children.stream().map(child -> child.symbol).collect(Collectors.toList());
		Production p = new Production((NonTerminal) symbol, rhs);
		if (!grammar.productions.contains(p))
			throw new IllegalArgumentException("No such production: " + p);
	}
		
	
	public DerivationTree(ContextFreeGrammar g, Symbol symbol, DerivationTree... children) {
		this(g, symbol, List.of(children));
	}
	
	public DerivationTree(ContextFreeGrammar g, Symbol symbol, List<DerivationTree> children) {
		this.grammar = g;
		this.symbol = symbol;
		this.children = Collections.unmodifiableList(children);
		check();
	}


	private static <T extends Comparable<? super T>> int compareLexicographic(Iterator<? extends T> left,
			Iterator<? extends T> right) {
		while (left.hasNext() && right.hasNext()) {
			int compare = left.next().compareTo(right.next());
			if (compare != 0) {
				return compare;
			}
		}
		if (left.hasNext()) {
			return 1;
		}
		if (right.hasNext()) {
			return -1;
		}
		return 0;
	}

	@Override
	public int compareTo(DerivationTree o) {
		int cmp = symbol.compareTo(o.symbol);
		if (cmp != 0)
			return cmp;
		return compareLexicographic(children.iterator(), o.children.iterator());
	}


	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((children == null) ? 0 : children.hashCode());
		result = prime * result + ((symbol == null) ? 0 : symbol.hashCode());
		return result;
	}


	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		DerivationTree other = (DerivationTree) obj;
		if (children == null) {
			if (other.children != null)
				return false;
		} else if (!children.equals(other.children))
			return false;
		if (symbol == null) {
			if (other.symbol != null)
				return false;
		} else if (!symbol.equals(other.symbol))
			return false;
		return true;
	}
	
	@Override
	public String toString() {
	    StringBuilder sb = new StringBuilder();
	    buildString(sb, "", true, true);
	    return sb.toString();
	}

	private void buildString(StringBuilder sb, String prefix, boolean isTail, boolean isFirst) {
		if (isFirst) {
			sb.append(symbol).append(System.lineSeparator());
		} else {
	        sb.append(isTail ? "└" : "├").append(symbol).append(System.lineSeparator());
	        prefix = prefix + (isTail ? " " : "│");
		}

	    for (int i = 0; i < children.size() - 1; i++) {
	        sb.append(prefix);
	        children.get(i).buildString(sb, prefix, false, false);
	    }
	    if (children.size() >= 1) {
	        sb.append(prefix);
	        children.get(children.size() - 1).buildString(sb,prefix, true, false);
	    }
	}

}
